{
  Solutions to the Advent Of Code.
  Copyright (C) 2023  Stefan Müller

  This program is free software: you can redistribute it and/or modify it under
  the terms of the GNU General Public License as published by the Free Software
  Foundation, either version 3 of the License, or (at your option) any later
  version.

  This program is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along with
  this program.  If not, see <http://www.gnu.org/licenses/>.
}

unit UGearRatiosTestCases;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, fpcunit, testregistry, USolver, UBaseTestCases, UGearRatios;

type

  { TGearRatiosFullDataTestCase }

  TGearRatiosFullDataTestCase = class(TEngineBaseTest)
  protected
    function CreateSolver: ISolver; override;
  published
    procedure TestPart1;
    procedure TestPart2;
  end;

  { TGearRatiosExampleTestCase }

  TGearRatiosExampleTestCase = class(TExampleEngineBaseTest)
  protected
    function CreateSolver: ISolver; override;
  published
    procedure TestPart1;
    procedure TestPart2;
  end;

  { TGearRatiosTestCase }

  TGearRatiosTestCase = class(TSolverTestCase)
  protected
    function CreateSolver: ISolver; override;
  published
    procedure TestEndOfLineNumber;
  end;

implementation

{ TGearRatiosFullDataTestCase }

function TGearRatiosFullDataTestCase.CreateSolver: ISolver;
begin
  Result := TGearRatios.Create;
end;

procedure TGearRatiosFullDataTestCase.TestPart1;
begin
  AssertEquals(530495, FSolver.GetResultPart1);
end;

procedure TGearRatiosFullDataTestCase.TestPart2;
begin
  AssertEquals(80253814, FSolver.GetResultPart2);
end;

{ TGearRatiosExampleTestCase }

function TGearRatiosExampleTestCase.CreateSolver: ISolver;
begin
  Result := TGearRatios.Create;
end;

procedure TGearRatiosExampleTestCase.TestPart1;
begin
  AssertEquals(4361, FSolver.GetResultPart1);
end;

procedure TGearRatiosExampleTestCase.TestPart2;
begin
  AssertEquals(467835, FSolver.GetResultPart2);
end;

{ TGearRatiosTestCase }

function TGearRatiosTestCase.CreateSolver: ISolver;
begin
  Result := TGearRatios.Create;
end;

procedure TGearRatiosTestCase.TestEndOfLineNumber;
begin
  FSolver.Init;
  FSolver.ProcessDataLine('...$541');
  FSolver.Finish;
  AssertEquals(541, FSolver.GetResultPart1);
end;

initialization

  RegisterTest('TGearRatios', TGearRatiosFullDataTestCase);
  RegisterTest('TGearRatios', TGearRatiosExampleTestCase);
  RegisterTest('TGearRatios', TGearRatiosTestCase);
end.
